"""
author: Daniel Riegelhaupt

this is a simple code generator for electrical circuits in Modelica
it takes an atom3 model and generates the appropriate code.

the following pieces are supported (for the moment ?)

from library Modelica.Electrical.Analog.Basic:
    Resistor
    Inductor
    Ground
    Capacitor

from library Modelica.Electrical.Analog.Sources:
    ConstantVoltage
    StepVoltage
    SineVoltage
"""

import tkMessageBox #needed to display infor/error message
import tkFileDialog #needed to pick the file we are going to generate the code in

SOURCE_IMPORT = "Modelica.Electrical.Analog.Sources"
BASIC_IMPORT = "Modelica.Electrical.Analog.Basic"

#### HELPERS ####

#retruns true if piece is a pos or negative gate
def isGate(piece):
    className = piece.__class__.__name__
    return (className == "p_gate" or className =="n_gate")

#returns true if the piece is a source
def isSource(piece):
    className = piece.__class__.__name__
    return (className == "SineVoltage" or className =="ConstantVoltage" or className =="StepVoltage")

#retuns true if the piece is a 2 gate basic piece
def isBasic(piece):
    className = piece.__class__.__name__
    return (className == "Capacitor" or className =="Inductor" or className =="Resistor")

#returns true if the piece is a one gate piece (for the moment this means only ground)
def isOneGate(piece):
    className = piece.__class__.__name__
    return (className == "Ground")


#gets the object whose out connection is the given neg gate
def getObjAttachedToN(gate):
    retObj = None
    assocs = gate.in_connections_
    for assoc in assocs:
        objs = assoc.in_connections_
        for obj in objs:
            if isBasic(obj) or isSource(obj):
                return obj #there can be only one object attached to it so return it
    return retObj #in case only another gate was atached to it or nothing at all

#gets the object whose in connection is the given pos gate
def getObjAttachedToP(gate):
    retObj = None
    assocs = gate.out_connections_
    for assoc in assocs:
        objs = assoc.out_connections_
        for obj in objs:
            if isBasic(obj) or isSource(obj)or isOneGate(obj):
                return obj #there can be only one object attached to it so return it
    return retObj #in case only another gate was atached to it or nothing at all

#get the neg gate attached to a piece
def getNGate(piece):
    gate = None
    if len(piece.out_connections_) >0:
        out =  piece.out_connections_[0]
        if len(out.out_connections_) >0 :
            gate = out.out_connections_[0] #no need to check for type it can only be a N gate
    return gate
    
#get the pos gate attached to a piece
def getPGate(piece):
    gate = None
    if len(piece.in_connections_) >0:
        i =  piece.in_connections_[0]
        if len(i.in_connections_) >0 :
            gate = i.in_connections_[0] #no need to check for type it can only be a P gate
    return gate

### get list of all posbile combinations of gates this is usefull when conencting beacause we need to now the gates ###

#get the pieces this pieces is conencted to from a begin gate (his nor p gate)
#to an end gate (the others p or n gate)
def getNextSelective(piece, getGateFun, otherGateName, getObjFromOtherGateFun):
    list = []
    #get the n gate
    sourceGate = getGateFun(piece)
    if sourceGate != None:
        assocs = sourceGate.out_connections_
        #for every other gate the n gate is conencted to
        for assoc in assocs:
            gates = assoc.out_connections_
            for gate in gates:
                #if this gate is also a neg gate
                if gate.__class__.__name__ == otherGateName:
                   retPiece = getObjFromOtherGateFun(gate)
                   #if this gate is attached to an object
                   if retPiece != None:
                       #add it to the list
                       list.append(retPiece)
    return list


#get a list of pieces reachable from this piece through its neg gate to a neg gate
def getNtoN(piece):
    return getNextSelective(piece, getNGate, "n_gate", getObjAttachedToN)

#get a list of pieces reachable from this piece through its neg gate to a pos gate
def getNtoP(piece):
    return getNextSelective(piece, getNGate, "p_gate", getObjAttachedToP)

#get a list of pieces reachable from this piece through its pos gate to a pos gate
def getPtoP(piece):
    return getNextSelective(piece, getPGate, "p_gate", getObjAttachedToP)

#get a list of pieces reachable from this piece through its pos gate to a neg gate
def getPtoN(piece):
    return getNextSelective(piece, getPGate, "n_gate", getObjAttachedToN)

### get a lsit of next and previous gates this usefulle when checking because then the gates don't matter####

#get the next pieces that are connected to the piece through the selected gate(except the current piece in case of p gate)
#choose the gate with the getGAteFunction
def getNext(piece, getGateFun):
    list = []
    sourceGate = getGateFun(piece)
    if sourceGate != None:
        assocs = sourceGate.out_connections_
        for assoc in assocs:
            gates = assoc.out_connections_
            for gate in gates:
                retPiece = None
                if gate.__class__.__name__ == "p_gate":
                    retPiece = getObjAttachedToP(gate)
                elif gate.__class__.__name__ == "n_gate":
                    retPiece = getObjAttachedToN(gate)
                else:
                    continue

                if retPiece != None:
                    list.append(retPiece)

    return list

#get the previous pieces that are connected to the piece through the selected gate(except the current piece in case of n gate)
#choose the gate with the getGateFunction
def getPrev(piece, getGateFun):
    list = []
    sourceGate = getGateFun(piece)
    if sourceGate != None:
        assocs = sourceGate.in_connections_
        for assoc in assocs:
            gates = assoc.in_connections_
            for gate in gates:
                retPiece = None
                if gate.__class__.__name__ == "p_gate":
                    retPiece = getObjAttachedToP(gate)
                elif gate.__class__.__name__ == "n_gate":
                    retPiece = getObjAttachedToN(gate)
                else:
                    continue

                if retPiece != None:
                    list.append(retPiece)

    return list

#get a list of pieces reachable from the pos gate (no matter through which type of gate that piece is connected)
def getNextP(piece):
    return getNext(piece, getPGate)

#get a list of pieces reachable from the neg gate (no matter through which type of gate that piece is connected)
def getNextN(piece):
    return getNext(piece, getNGate)

#get a list of pieces that are conencted to this piece
#in other words pieces who have arrows going from them to this piece through the pos gate
def getPrevP(piece):
    return getPrev(piece, getPGate)

#get a list of pieces that are conencted to this piece
#in other words pieces who have arrows going from them to this piece through the neg gate
def getPrevN(piece):
    return getPrev(piece, getNGate)

#### END HELPERS ####


"""
the constraint checker class: will check if a model is valid or not

the costraints are:
*each circuit must be closed
*each circuit must contain at least 1 source
*each circuit must contain exactly 1 ground
"""
#TODO make sure there is no closed crircuit without a source and one with a source in it
#in other words check that the closed circuit has a source not that there is a source in the model
#TODO implement this class
class ConstraintChecker:
    def __init__(self, atom3i, ASGroot):
        self.atom3i =atom3i
        self.root = ASGroot
        self.listSources = set() #list of sources sources
        self.listBasic = set() #pieces with 2 gates (excpet sources)
        self.listOneGate = set() #only grounds
        self.listAll = set() #when checking everyhting it is quicker this way

        counter = 0
        nodeTypeList = self.root.listNodes.keys()
        for nodeType in nodeTypeList:
            nodeList = self.root.listNodes[nodeType]
            for node in nodeList:
                if isSource(node):
                    node.reachedBySource = False #set a bool that tells if a piece is reachacle by a source
                    node.index = counter #the index of the source so we now to which source a ground belongs
                    counter = counter +1
                    node.listOfOtherSources = set() #list of other sources that can reach this needed for ground correction procedure
                    self.listSources.add(node)
                    self.listAll.add(node)
                elif isOneGate(node):
                    node.reachedBySource = False
                    node.listOfSources = set() 
                    """"
                    a list containing the inces of the sources that can reach it (that are in the same circuit)
                    *those list must be mutely disjunct !!! or else there are more then one ground on the same circuit
                    *the sum of the len of the this lsit must be equal tot the number of sources
                    if not there is a circuit without a ground
                    """
                    self.listOneGate.add(node)
                    self.listAll.add(node)
                elif isBasic(node):
                    node.index = -1 #source index that can reaches it first
                    node.reachedBySource = False
                    self.listBasic.add(node)
                    self.listAll.add(node)
                else: #its a gate ignore
                    continue

    # check if every things is connected
    def __isConnected(self):
        #for all objects
        for obj in self.listAll:
            if obj in self.listOneGate:
                #for one gate it doesn't matter if the arrows arrive or leaves from the p gate
                #so only false if both are empty
                if len(getNextP(obj)) == 0 and len(getPrevP(obj)) == 0:
                    return False
            else: #it has two gates
                #both sides must be connected to other pieces it can either be incomming or outgoing
                if (len(getNextP(obj)) == 0 and len(getPrevP(obj)) == 0) or\
                (len(getNextN(obj)) == 0 and len(getPrevN(obj)) == 0):
                    return False
        return True #will only return true if everything is connected


    def __getReachableHelper(self, list, sourceIndex, bool):
        for elem in list:
            if elem.reachedBySource == False:
                elem.reachedBySource = True
                if isOneGate(elem):
                    elem.listOfSources.add(sourceIndex)
                elif isBasic(elem):
                    elem.index = sourceIndex
                elif isSource(elem):
                    elem.listOfOtherSources.add(sourceIndex)

                self.__getReachable(sourceIndex, elem)
            elif bool: #if the element we are in is reachable from another source and the current piece is a source
                #they belong to the same circuit therefore add it to the list of sources assoiated to the ground
                index = elem.index
                for ground in self.listOneGate:
                    if not(sourceIndex in ground.listOfSources) and (index in ground.listOfSources):
                        ground.listOfSources.add(sourceIndex)
            else:
                continue

    def __isWorthChecking(self, list):
        if len(list) > 0:
            #check to see if the lsit contains at least one lement that hasn't been reached
            for elem in list:
               if elem.reachedBySource == False:
                   return True #if an element that wasn't reached was found we must check the list
            return False
        else:
            return False

    #check recursuvely for all pieces that can be reached transversaly through the given piece
    def __getReachable(self,sourceIndex ,piece):
        listP = getNextP(piece)
        listN = getNextN(piece)
        listPrevP = getPrevP(piece)
        listPrevN = getPrevN(piece)
        
        bool = isSource(piece) #if the piece is a source

        if self.__isWorthChecking(listP):
            self.__getReachableHelper(listP, sourceIndex, bool)
        if self.__isWorthChecking(listN):
            self.__getReachableHelper(listN, sourceIndex, bool)
        if self.__isWorthChecking(listPrevP):
            self.__getReachableHelper(listPrevP, sourceIndex, bool)
        if self.__isWorthChecking(listPrevN):
            self.__getReachableHelper(listPrevN, sourceIndex, bool)

    #depending on the order in wich the sources are visited
    #the list of sources in the same circuit of the ground may be incomplete (simply beacuase the ground was visited after the source)
    #in here we take corrective measures
    def __groundCorrection(self):
        if len(self.listSources) > 0:
            for source in self.listSources:
                #if a source is reachable from other sources
                if len(source.listOfOtherSources) > 0:
                    #for every ground
                    for ground in self.listOneGate:
                        #if the current ground doen't have this source in the list of reachable sources
                        if not (source.index in ground.listOfSources):
                            #if a source that reacehs the current source also reacehs the ground add it else don't
                            if len(source.listOfOtherSources.intersection(ground.listOfSources)) > 0:
                               ground.listOfSources.add(source.index)


    def __hasSources(self):
        #check if there are any sources
        if len(self.listSources) > 0:
            for source in self.listSources:
                if not source.reachedBySource: #if the source hasn't already been checked while cheking another source
                    source.reachedBySource = True #it is a source itself
                    self.__getReachable(source.index, source)

            #we coorect the list of reachable sources by the ground that  may not have been completed during the normal run
            self.__groundCorrection()

            #we now know for evey piece wheter it belongs to a cicuit that has a source or not
            #we know check this for every piece
            for elem in self.listAll:
                #if there is still an elment that isn't reacable that means it belongs to a cicuit without a source
                if elem.reachedBySource == False:
                   return False
            return True #if every element was reached then every circuit has at least on source
        else:
            return False

    def __hasGround(self):
        if len(self.listOneGate) > 0:
            total = 0
            #check that the list of sources are mutualt disjunct
            for ground in self.listOneGate:
                for ground2 in self.listOneGate:
                    if ground != ground2:
                        inter = ground.listOfSources.intersection(ground2.listOfSources)
                        if len(inter) != 0:
                            return False #there are more then one ground on the same circuit
                total = total + len(ground.listOfSources)
            """
            if they are mutualy disjunct check if every circuit has ground
            by taking the total of number of sources all grounds can reach and seing if it is equal to 
            the number of sources there are.(this works beacause we already know each circuit has at least one source and no ground has a source in common)     
            """
            if total == len(self.listSources):
                return True
            else:
                return False
        else:
            return False


    def check(self):
        if self.__isConnected():
            if self.__hasSources():
                if self.__hasGround():
                    return (True, "the model is valid")
                else:
                    return (False, "Every circuit must have exactly 1 ground")
            else:
                return (False, "Every circuit must have at least 1 source")
        else:
            return (False, "Every circuit must be closed!")
      

"""
The code generator class: it generates modelica code from the given atom3 model
the only method a user should use is the generate method.
IMPORTANT: when using the generate method the PRECONDITION is to have a valid model
(a model is valid if the ConstraintChecker.check method returns true)
"""

class Generator:
    def __init__(self, atom3i, ASGroot, fileName, modelName):
        self.atom3i = atom3i
        self.root = ASGroot
        self.file = open(fileName, 'w') #open file in write mode, existing file with this name will be overwritten
        self.modelName = modelName

    """
    PUBLIC METHODS
    """
    """
    generates the modelica code
    PRECONDITION: the atom3 model must be valid (use the ConstraintChecker.check method to check this)
    """
    def generate(self):
        self.file.write("model " + self.modelName + "\n")
        self.__writeDeclarations()
        self.file.write("equation\n")
        self.__writeConnections()
        self.file.write("end " + self.modelName + ";")
        self.file.close()

    """
    PRIVATE METHODS
    """
    #returns the attribues EXCPET NAME !!!! of a piece
    def __getAttributes(self, piece):
        str = ""
        keys = piece.generatedAttributes.keys()
        counter = 0
        for key in keys:
            if key != "name":
                if counter > 0:
                    str = str + ", "
                str = str + key + " = " + piece.getAttrValue(key).toString()
                counter = counter + 1
        return str
    

    #a helper function does the actual declaring
    def __declareHelper(self, importer, node):
        self.file.write("\t"+ importer +"." + node.__class__.__name__ + " " + node.name.getValue())
        str = self.__getAttributes(node)
        attr = ""
        if str != "":
            attr = '('+ str +')'
        self.file.write(attr + ";\n")

    def __writeDeclarations(self):
        nodeTypeList = self.root.listNodes.keys()
        for nodeType in nodeTypeList:
            nodeList = self.root.listNodes[nodeType]
            for node in nodeList:
                if isSource(node):
                    self.__declareHelper(SOURCE_IMPORT, node)
                    ("(V = " + str(node.V.getValue())+");\n")
                elif isOneGate(node) or isBasic(node) :
                    self.__declareHelper(BASIC_IMPORT, node)
                else:
                    continue

    #does the actual file writting of the connections
    def __connectHelper(self, node ,list, ownGate, otherGate):
        if len (list) > 0:
            for piece in list:
                self.file.write("\tconnect("+node.name.getValue()+"."+ownGate)
                self.file.write(", " + piece.name.getValue() + "." + otherGate +");\n")

    #write the connections for this piece
    def __connect(self, node):
        self.__connectHelper(node, getNtoN(node), 'n', 'n')
        self.__connectHelper(node, getNtoP(node), 'n', 'p')
        self.__connectHelper(node, getPtoN(node), 'p', 'n')
        self.__connectHelper(node, getPtoP(node), 'p', 'p')

    def __writeConnections(self):
        nodeTypeList = self.root.listNodes.keys()
        for nodeType in nodeTypeList:
            nodeList = self.root.listNodes[nodeType]
            for node in nodeList:
                if isSource(node) or isOneGate(node) or isBasic(node):
                    self.__connect(node)
                else:
                    continue

"""
The procedure that needs to be called in atom3
parameters:
    * atom3i: the atom3 interface
    * metaModelName: the name of the metaModel, this should normaly never change
"""
def generate(atom3i, metaModelName = 'Electrical_META' ):
    #get the root
    ASGroot = atom3i.ASGroot.getASGbyName(metaModelName)
    #check the constraints
    checker = ConstraintChecker(atom3i, ASGroot)
    (result, errorMsg) = checker.check()
    if result:
        #get the name of the model
        modelName = ASGroot.name.getValue()
        if modelName == "":
            modelName = "NoName"
        #get the name of the file.
        fileName = tkFileDialog.asksaveasfilename(title='Save file as',initialfile= modelName\
        + '.mo' ,defaultextension='.mo',filetypes=[('modelica files', '*.mo')])
        if fileName: #if there is no file name cancel was pressed
            print fileName
            #generate the modelica code
            gen = Generator(atom3i, ASGroot, fileName, modelName)
            gen.generate()
    else:
        tkMessageBox.showerror("Constraint not met!", errorMsg)